/*
 * Soubor:  	trace.c
 * Datum:   	26.09.2012
 * Autor:   	Tomas Valek, xvalek02@stud.fit.vutbr.cz
 * Popis:   	Trace. Program umoznuje pruchod paketu siti k objeveni
				smerovacu. Podpora IPv4/IPv6. Pouziva ICMPv4/ICMPv6 zpravy.
 */

/*POZNAMKY:

	struct addrinfo {
		int              ai_flags;
		int              ai_family;
		int              ai_socktype;
		int              ai_protocol;
		size_t           ai_addrlen;
		struct sockaddr *ai_addr;
		char            *ai_canonname;
		struct addrinfo *ai_next;
	}

	struct sockaddr {
        unsigned short sa_family;			2 bajty
        char sa_data[14];					14 bajtu
	};

	IPv4 socket
	struct sockaddr_in {
        short 			sin_family;			2 bajty		AF_INET, AF_INET6
        unsigned short 	sin_port;			2 bajty		htons(3490)
        struct in_addr 	sin_addr;			4 bajty		see struct in_addr, below
        char 			sin_zero[8];		8 bajtu		zero this if you want to
	};

	sockaddr zabira 16 bajtu, sockaddr_in zabira 16 bajtu-lze je tedy pretypovavat

	struct in_addr {
    	unsigned long s_addr;          		4 bajty		load with inet_pton()
	};

	IPv6 socket
	struct sockaddr_in6 {
		sa_family_t     sin6_family;    AF_INET6
		in_port_t       sin6_port;      port number
		uint32_t        sin6_flowinfo;  IPv6 flow information
		struct in6_addr sin6_addr;      IPv6 address
		uint32_t        sin6_scope_id;  Scope ID (new in 2.4)
	};

	struct in6_addr {
		unsigned char   s6_addr[16];   IPv6 address
	};

	struct icmphdr
	{
 		 u_int8_t type;		 message type 
 		 u_int8_t code;		 type sub-code 
 		 u_int16_t checksum;
 		 union
 		 {
			struct
			{
	  			u_int16_t	id;
	  			u_int16_t	sequence;
			} echo;			 		 echo datagram 
			u_int32_t	gateway;	 gateway address 
			struct
			{
	  			u_int16_t	__unused;
	  			u_int16_t	mtu;
			} frag;			 path mtu discovery 
  		} un;
	};

	struct icmp6_hdr {
	u_int8_t	icmp6_type;	 	type field 
	u_int8_t	icmp6_code;	 	code field 
	u_int16_t	icmp6_cksum;	checksum field 
		union {
			u_int32_t	icmp6_un_data32[1]; 	 type-specific field 
			u_int16_t	icmp6_un_data16[2]; 	 type-specific field 
			u_int8_t	icmp6_un_data8[4];  	 type-specific field 
		} icmp6_dataun;
	};

	#define INET_ADDRSTRLEN 16
	#define INET6_ADDRSTRLEN 46

 	struct iphdr {
         __u8    version:4,
                 ihl:4;
         __u8    tos;
         __u16   tot_len;
         __u16   id;
         __u16   frag_off;
         __u8    ttl;
         __u8    protocol;
         __u16   check;
         __u32   saddr;
         __u32   daddr;
         The options start here.
	 };
*/

#include <stdio.h>				/*vstup, vystup*/
#include <stdlib.h>				/*obecne funkce jazyka C, EXIT_SUCCESS a EXIT_FAILURE*/
#include <string.h> 			/*strcmp*/
#include <getopt.h>
#include <stdbool.h> 			/*true, false*/
#include <netdb.h>				/*getaddrinfo()*/			/*lze pouzit*/
#include <arpa/inet.h>			/*inet_ntop*/				/*lze pouzit*/
#include <sys/socket.h>
#include <netinet/ip_icmp.h>	/*ICMPv4*/
#include <netinet/icmp6.h>		/*ICMPv6*/
#include <unistd.h>				/*getpid(), close()*/
#include <netinet/ip.h>			/*struktura iphrd IP protokol*/
#include <sys/time.h>			/*timeval*/
#include <signal.h>

#define NI_MAXHOST      1025
#define IHL				4		/*konstanta pro vypocet delky IPv4 hlavicky*/
#define IPV6_HEADER 	40 		/*hlavicka IPv6 ma pevnych 40bajtu*/
#define MIN_PARAM 		2 		/*min. pocet parametru*/
#define MAX_PARAM 		6		/*max. pocet parametru*/
#define BUFFER_SIZE 	1508	/*velikost bufferu pro prichozi zpravu Recvfrom*/
#define RANDOM_PORT 	0		/*port 0 znaci nahodny port*/
#define TIMEOUT			3		/*timeout pro cekani na odpoved je 3s. dle zadani*/

/*DEBUG:*/
#define ARRAY 0
#define DEBUG 0

/*Globalni promenne kvuli signalum*/
int sockfd = -1; 					/*navratova hodnota socketu(file deskriptor)*/
struct icmphdr *icmpSend4 = NULL;	/*ICMPv4 odchozi*/
struct icmp6_hdr *icmpSend6 = NULL; /*ICMPv6 odchozi*/
struct addrinfo *current = NULL;	/*pro getaddrinfo()*/

/*Struktury*/
typedef struct{
	int firstTTL;	/*velikost TTL pro prvni paket*/
	int maxTTL;		/*max. velikost TTL*/
	char* host;		/*IPv4/IPv6/DNS cil*/
}Tparameter;

/*Prototypy*/
int main( int argc, char* argv[] );
void getParams(int argc, char* argv[], Tparameter *param);
void error(int number,int sockfd, void* malloc);
void traceroute(Tparameter param, char* ipAdress, bool IPv6);
bool setTTL(int sockfd, int ttl, bool IPv6);
void getIP(char* host,char* ipAdress, bool* IPv6);
void getHostName(char* ipAdress, char* hostName,  bool IPv6);
unsigned short checksum(unsigned short *ip, int len);
bool lostMyPacket(char* recvBuffer, unsigned short ipHeaderLength, 
unsigned short id, unsigned short sequence, bool ttlNull, bool IPv6);
void terminated(int sigNumber);
double response(int timeA, int utimeA, int timeB, int utimeB);

/*Chybove stavy:*/
enum ERROR{

	ERR_PARAM = 0,
	ERR_SSCANF,
	ERR_TTL,
	ERR_MALLOC,
	ERR_INET_PTON,
	ERR_UNKNWN
};

char *errorMessages [] ={

	/*ERR_PARAM:*/
	"Chybne zadane parametry! Pro napovedu spustte program s prepinacem -h.\n",
	/*ERR_SSCANF*/
	"Chyba ve funkci SSCANF. Nezadali jste cislo nebo se nepodarila se konverze\
	 retezce na cislo. Pro napovedu spustte program s prepinacem -h.\n",
	/*ERR_TTL*/
	"firstTTL a maxTTL nemuze byt < 1.\n",
	/*ERR_MALLOC*/
	"Chyba pri alokaci dat(malloc).\n",
	/*ERR_INET_PTON*/
	"Sitova adresa neni platna.\n",
	/*ERR_UNKNWN*/
	"Neznama chyba, pri cteni paketu.\n.",
};

/*Napoveda:*/
char help[] =
"Program trace.\n"
"Autor: Tomas Valek.\n"
"Datum: 26.09.2012.\n"
"Program umoznuje pruchod paketu siti k objeveni smerovacu.\n"
"Pouziti: trace [-f velikost_TTL_pro_paket][-m max_TLL] host\n"
"-f: velikost TTL pro prvni paket-implicitne 1\n"
"-m: max. velikost TTL-implicitne 30\n"
"host: povinny parametr, IPv4/IPv6/DNS cil.\n";

int main( int argc, char* argv[] ){
/*Hlavni funkce programu trace.*/

	char ipAdress[INET6_ADDRSTRLEN+1]; 	/*pole do ktereho se uklada textova IP adresa + '\0' */
	bool IPv6 = false;

	Tparameter param;
	param.firstTTL = 1;			/*implicitni hodnota podle zadani*/
	param.maxTTL = 30;			/*implicitni hodnota podle zadani*/
	param.host = NULL;

	getParams(argc, argv, &param);

	/*Kontrola parametru:
	printf("param.firstTTL:\t%d\n",param.firstTTL);
	printf("param.maxTTL:\t%d\n",param.maxTTL);
	printf("param.host:\t%s\n",param.host);*/

	signal(SIGINT, terminated);
	signal(SIGHUP, terminated);

	getIP(param.host, ipAdress, &IPv6);

	traceroute(param, ipAdress, IPv6);

	return EXIT_SUCCESS;
}

double response(int timeA, int utimeA, int timeB, int utimeB){
/*Vypocita odezvu*/

	/*printf("novy cas:%.3f\n",(utimeB-utimeA)/1000.0);*/
	return (timeB-timeA)*1000+(utimeB-utimeA)/1000.0;
}

void terminated(int sigNumber){
/*Oblouzeni signalu*/

	if (sockfd != -1)
		close(sockfd);

	if (icmpSend4 != NULL)
		free(icmpSend4);

	if (icmpSend6 != NULL)
		free(icmpSend6);

	if (current != NULL)
		freeaddrinfo(current);

	++sigNumber;/*Aby nehlasil prekladac varovani*/

	exit(0);
}

void getHostName(char* ipAdress, char* hostName,  bool IPv6){
/*Ziska jmeno hosta z IP adresy. Pouziva novejsi funkci getnameinfo, ktera
 nahrazuje gethostbyaddr. Pokud jmeno neziska nebo nastane chyba, nastavi IP
 adresu jako jmeno hosta.
ipAdress	IP adresa
hostName	zde bude ulozeno jmeno hostitele
V pripade neuspechu vraci funkce false.
*/

	struct sockaddr_in6 host6;
	struct sockaddr_in host;

	int r;
	int s;

	if(IPv6){
		host6.sin6_family = AF_INET6;
		host6.sin6_flowinfo = 0;/*init*/
		host6.sin6_scope_id = 0;/*init*/
		s = inet_pton(AF_INET6, ipAdress, &(host6.sin6_addr));
	}
	else{
		host.sin_family = AF_INET;
		s = inet_pton(AF_INET, ipAdress, &(host.sin_addr));
	}

	if (s <= 0){
		if (s == 0)
			fprintf(stderr, "%s", errorMessages[ERR_INET_PTON]);
		else
			perror("inet_pton()");

		return;
	}
 
	if(IPv6)
		r = getnameinfo((struct sockaddr*)&host6, sizeof(host6), hostName, NI_MAXHOST, NULL, 0, 0);
	else
		r = getnameinfo((struct sockaddr*)&host, sizeof(host), hostName, NI_MAXHOST, NULL, 0, 0);

	if (r){/*getnameinfo nastavuje jako uspech 0*/
		printf("getnameinfo():%s\n", gai_strerror(r));
		memcpy(hostName, ipAdress, strlen(ipAdress));
	}
	
	/*printf("hostName:%s\n", hostName);*/
}

void getIP(char* host,char* ipAdress, bool* IPv6){
/*Ziska IPvX adresu*/

	struct addrinfo hints;
#if ARRAY
	struct addrinfo *result;
#endif
	/*struct addrinfo *current;*/ /*Globalni promenna*/

	int s;
	void *ptr;

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_CANONNAME;	/*nazev, napr. www.seznam.cz*/

	s = getaddrinfo(host, NULL, &hints, &current);

	if (s != 0){/*chyba getaddrinfo()*/
		/*gai_strerror() prevadi chybovy kod na retezec*/
		fprintf(stderr, "getaddrinfo(): %s\n", gai_strerror(s));
		exit(-1);
	}

#if DEBUG
	printf ("Host: %s\n", host);
	printf("Delka IP adresy(znaky):%d\n", current->ai_addrlen);
#endif

#if ARRAY /*zapnout v pripade ze chceme vypsat cele pole*/
	for(current = result; current != NULL; current = current->ai_next) {
#endif

		switch (current->ai_family){
			/*podle typu IPvX. Preferujeme IPv6 adresu*/
			case AF_INET6:
				/*ai_addr pretypujeme na struct sockaddr_in viz komentar vyse*/
				ptr = &((struct sockaddr_in6 *) current->ai_addr)->sin6_addr;
				*IPv6 = true;
				break;

			case AF_INET:
				ptr = &((struct sockaddr_in *) current->ai_addr)->sin_addr;
				break;
#if ARRAY
			default:
				continue;
#endif
		}

		/*inet_ntop - prevedeme IPv4 a IPv6 adresu z binarni formy do textove*/
		inet_ntop (current->ai_family, ptr, ipAdress, INET6_ADDRSTRLEN+1);

		/*vypis:
		if(current->ai_family == AF_INET6)
			printf("IPv6:\t%s\t%s\n", ipAdress, current->ai_canonname);
		else
			printf("IPv4:\t%s\t%s\n", ipAdress, current->ai_canonname);*/

#if ARRAY
	}
#endif

	freeaddrinfo(current);
	current = NULL;
}

void traceroute(Tparameter param, char* ipAdress, bool IPv6){
/*Traceroute, v teto funkci pouzivam globalni promenne.
Algoritmus:

	1) nastavime hodnotu TTL implicitne na 1, popr. dle uzivatele param.(first_ttl)
	2) odesleme ECHO_REQUEST
	3) cekame na prichozi ICMP paket nez vyprsi TIMEOUT, zajimaji nas pouze tri typy:
		a)ECHOREPLY- vypis posledniho a jdi na bod 8
		b)ztrata naseho paketu-vypis informace o smerovaci na kterem byl zahozen,
		 sequence++, ttl++ a jdi na bod 2
		c)chyby dane zadanim- vypis informace o smerovaci na kterem nastala chyba,
		 a jdi na bod 8
	7a) pokud je TTL > max hodnota TTL(parametr max_ttl) tak konec
	7b) pokud je TTL <= max hodnota TTL(parametr max_ttl) pokracuj na bod 2
	8) Konec
*/

	/*Globalni promenne v teto funci:
	int sockfd = -1; 					navratova hodnota socketu(file deskriptor)
	struct icmphdr *icmpSend4 = NULL;	ICMPv4 odchozi
	struct icmp6_hdr *icmpSend6 = NULL;	ICMPv6 odchozi
	*/

	fd_set mySet;				/*mnozina(bitove pole) slouzi k uchovani file deskriptoru*/

	/*ICMP:*/
	struct icmphdr *icmpRecv4;			/*ICMPv4 prichozi*/
	struct icmp6_hdr *icmpRecv6;		/*ICMPv6 prichozi*/
	uint16_t pid;						/*ID procesu, patri do hlavicky ICMP*/
	uint16_t sequence = 1;				/*poradove cislo zadosti, patri do hlavicky ICMP*/

	/*Sokety*/
	struct sockaddr_in sendSock4;		/*IPv4 socket-odchozi*/
	struct sockaddr_in receiveSock4;	/*IPv4 socket-prichozi*/
	struct sockaddr_in6 sendSock6;		/*IPv6 socket-odchozi*/
	struct sockaddr_in6 receiveSock6;	/*IPv6 socket-prichozi*/
	socklen_t SockLength;				/*velikost socketu*/

	struct iphdr *ip;					/* hlavicka IPv4 protokolu */

	/*timeout+odezva*/
	struct timeval tv;					/*select-timeout*/
	struct timeval ping;				/*odezva*/
	int timeA;
	int utimeA;
	int timeB;
	int utimeB;

	char recvBuffer[BUFFER_SIZE];	/*buffer pro prichozi zpravu Recvfrom*/
	int recvLength;					/*pocet prijatych bajtu z funkce recvfrom*/
	char ipAdressString[INET6_ADDRSTRLEN+1];	/*IP adresa v teckove notaci*/
	char hostName[NI_MAXHOST];		/*pole pro nazev hostitele*/
	int sendOfChar;					/*pocet odeslanych znaku z funkce sendto*/
	bool lostMyPacketRet;			/*navratova hodnota pro lostMyPacket()*/
	bool ttlNull;					/*indikuje TTL=0*/
	bool echoReply;					/*indikuje echoReply*/
	bool host;
	bool network;
	bool protocol;
	bool communication;

	/*Ukazatele na ICMPv4 nebo ICMPv6, kterou budeme odesilat*/
	size_t size_icmp_hdr46 = 0;
	u_int8_t* icmp_code46 = NULL;
	u_int16_t* icmp_id46 = NULL;
	u_int16_t* icmp_cksum46 = NULL;
	u_int16_t* icmp_sequence46 = NULL;
	/*Ukazatele na ICMPv4 nebo ICMPv6, kterou prijmeme*/
	u_int16_t* icmp_idSend46 = NULL;
	u_int16_t* icmp_sequenceSend46 = NULL;

	/*Slouzi pro ukazatel na alokovanou pamet pro ICMPv4/ICMPv6*/
	void* mallocPtr46 = NULL;

	/*Nastavime polozky souvisejici s IPv6/IPv4 + ukazatele pro IPv6/IPv4*/
	if(IPv6){
		size_icmp_hdr46 = sizeof(struct icmp6_hdr);

		icmpSend6 = (struct icmp6_hdr *)malloc(size_icmp_hdr46);
		mallocPtr46 = icmpSend6;
		if (icmpSend6 == NULL)/*Chyba malloc*/									/*otestovano valgrindem*/
			error(ERR_MALLOC, -1, NULL);

		icmpSend6->icmp6_type = ICMP6_ECHO_REQUEST;

		/*nastaveni ukazatelu pro ICPMv6*/
		icmp_code46 = &(icmpSend6->icmp6_code);
		icmp_id46 = &(icmpSend6->icmp6_dataun.icmp6_un_data16[0]);		/*ID*/
		icmp_cksum46 = &(icmpSend6->icmp6_cksum);
		icmp_sequence46 = &(icmpSend6->icmp6_dataun.icmp6_un_data16[1]);/*SEQUENCE*/
	}
	else{
		size_icmp_hdr46 = sizeof(struct icmphdr);

		icmpSend4 = (struct icmphdr *)malloc(size_icmp_hdr46);
		mallocPtr46 = icmpSend4;
		if (icmpSend4 == NULL)													/*otestovano valgrindem*/
			error(ERR_MALLOC, -1, NULL);

		icmpSend4->type = ICMP_ECHO;

		icmp_code46 = &(icmpSend4->code);
		icmp_id46 = &(icmpSend4->un.echo.id);
		icmp_cksum46 = &(icmpSend4->checksum);
		icmp_sequence46 = &(icmpSend4->un.echo.sequence);
	}


	/*vytvorime socket
		AF_INET = Address Family, Internet = IP Addresses
		PF_INET = Packet Format, Internet = IP, TCP/IP or UDP/IP
		AF/PF je to jedno, konstanty jsou stejne
        SOCK_RAW     	typ soketu(mame pod kontrolou tvorbu vsech hlavicek)
        IPPROTO_ICMP	protokol ICMP
	POZOR! aby se RAW socket vytvoril, je treba byt ROOT*/
	if(IPv6)
		sockfd = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
	else
		sockfd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);

	if ( sockfd == -1 )
	{/*socket se nepodarilo vytvorit*/
		perror("Chyba pri vytvareni socketu:");									/*otestovano valgrindem*/
		error(-1, -1, mallocPtr46);
	}

	pid = getpid();

	/*nastaveni ICMP-CODE a ICMP-ID*/
	*icmp_code46 = htons(0);
	*icmp_id46 = htons(pid); /*naplnime cislem procesu-jednoznacny identifikator*/

	/*naplneni struktury sockaddr_in a sockaddr6_in*/
	if(IPv6){
		sendSock6.sin6_family = AF_INET6;
		sendSock6.sin6_port = htons(RANDOM_PORT);
		sendSock6.sin6_flowinfo = 0;/*init*/
		sendSock6.sin6_scope_id = 0;/*init*/
		inet_pton (AF_INET6, ipAdress,  &sendSock6.sin6_addr);
	}
	else{
		sendSock4.sin_family = AF_INET;
		sendSock4.sin_port = htons(RANDOM_PORT);
		inet_pton (AF_INET, ipAdress,  &sendSock4.sin_addr);
	}

	do{

		/*nastaveni struktury timeval, musi byt inicializovany obe, select pouziva obe*/
		tv.tv_sec = TIMEOUT;/*max doba cekani na paket v (s.)*/
		tv.tv_usec = 0;		/*mikrosekundy*/

		/*doplnujeme hlavicku ICMP v cyklu, protoze checksum se od sebe lisi*/
		*icmp_cksum46 = htons(0);
		/*protoze posilame vice ECHO zadosti-odlisime jeste poradovym cislem*/
		*icmp_sequence46 = htons(sequence);

		if ( setTTL(sockfd, param.firstTTL, IPv6) == false )
			error(-1, sockfd, mallocPtr46);										/*otestovano valgrindem*/

		if(IPv6){
			icmpSend6->icmp6_cksum = checksum((unsigned short *)icmpSend6, sizeof(struct icmp6_hdr));
			/*posleme ICMP zadost ECHO na soket. Pozn. neposilame IP hlavicku. 
				Buffer obsahuje pouze ICMP hlavicku.
				1.param			file deskriptor socketu
				2.param			ukazatel na buffer
				3.param			delka
				4.param			flags
				5.param			cil
				6.param			delka
			*/
			sendOfChar = sendto(sockfd, (char *)icmpSend6, sizeof(struct icmp6_hdr), 0, (struct sockaddr *)&sendSock6, sizeof(sendSock6));
		}
		else{
			icmpSend4->checksum = checksum((unsigned short *)icmpSend4, sizeof(struct icmphdr));

			sendOfChar = sendto(sockfd, (char *)icmpSend4, sizeof(struct icmphdr), 0, (struct sockaddr *)&sendSock4, sizeof(sendSock4));
		}

		if (sendOfChar == -1){
			perror("Chyba funkce sendto():");
			error(-1, sockfd, mallocPtr46);										/*otestovano valgrindem*/
		}

		/*ziskame casovy bod A*/
		gettimeofday(&ping, NULL);
		timeA = ping.tv_sec;
		utimeA = ping.tv_usec;

		do{/*prijimani vsech ICMP paketu*/

			/*INIT*/
			lostMyPacketRet = false;
			ttlNull = false;
			echoReply = false;
			host = false;
			network =  false;
			protocol = false;
			communication =  false;
			icmpRecv6 = NULL;
			icmpRecv4 = NULL;
			ip = NULL;

			/*Do mnoziny si ukladame odeslane file deskriptory*/
			FD_ZERO(&mySet);		/*vyprazdnime mnozinu, protoze mnozina obsahuje nahodna data*/
			FD_SET(sockfd, &mySet); /*vlozime file deskriptor do mnoziny*/

			/*select-prace s vice sokety
			1.param		file deskriptor soketu
			2.param		ukazatel na mnozinu s file deskriptorama z kterych chceme cist
			3.param		ukazatel na mnozinu soketu pro zapis
			4.param		ukazatel na mnozinu soketu u kterych chceme zjisti, zda nastala nejaka chyba
			5.param		ukazatel na strukturu-jak dlouho se ceka
			*/
			if (select(sockfd + 1, &mySet, NULL, NULL, &tv) == -1){
				perror("Chyba ve funkci select():");
				break;
			}

			if ( FD_ISSET(sockfd, &mySet) > 0 ){/*otestujeme zda je file deskriptor v mnozine*/

				/*recvfrom-cekame na zpravu
				1.param		socket
				2.param		buffer pro prichozi zpravu
				3.param		velikost bufferu
				4.param		flags
				5.param		cil
				6.param		velikost
				*/
				if(IPv6){
					SockLength = sizeof(receiveSock6);
					recvLength = recvfrom(sockfd, recvBuffer, BUFFER_SIZE, 0,(struct sockaddr *)&receiveSock6, &SockLength);
				}
				else{
					SockLength = sizeof(receiveSock4);
					recvLength = recvfrom(sockfd, recvBuffer, BUFFER_SIZE, 0,(struct sockaddr *)&receiveSock4, &SockLength);
				}

				/*nastavime casovy bod B*/
				gettimeofday(&ping, NULL);
				timeB = ping.tv_sec;
				utimeB = ping.tv_usec;

				if(recvLength == -1)/*chyba, ale program muze bezet dale*/
					perror("Chyba funkce recvfrom():");

				/*podle typu paketu:*/
				if(IPv6)
				{/*Pro ICMPv6 se nevraci paket s IP hlavickou jako je tomu u IPv4.
					v tele ICMPv6:
					ECHOREPL:	data
					TIME:		IP+ORIG datagram
					DESTINATION	IP+ORIG datagram
				 */

					/*pretypujeme na ICMPv6 hlavicku*/
					icmpRecv6 = (struct icmp6_hdr *)recvBuffer;

					switch (icmpRecv6->icmp6_type){
						case ICMP6_ECHO_REPLY: /*TYPE 129*/
							if (icmpRecv6->icmp6_code == 0)
								echoReply = true;
							else
								error(ERR_UNKNWN, sockfd, mallocPtr46);
							break;
						case ICMP6_TIME_EXCEEDED: /*TYPE 3*/
							if (icmpRecv6->icmp6_code == ICMP6_TIME_EXCEED_TRANSIT)	/*TTL = 0*/
								ttlNull = true;
							break;
						case ICMP6_DST_UNREACH:		/*TYPE 1*/
							/*communication administratively prohibited X!*/
							if ( icmpRecv6->icmp6_code == ICMP6_DST_UNREACH_ADMIN )
								communication = true;
							break;
					}

					if (echoReply){/*pri echoReply prijdou primo data*/
						icmp_idSend46 = &(icmpRecv6->icmp6_dataun.icmp6_un_data16[0]);
						icmp_sequenceSend46 = &(icmpRecv6->icmp6_dataun.icmp6_un_data16[1]);
					}
					else{/*Presuneme se na hlavicku ICMPv6, kterou jsme odesilali, je v tele prijate ICMPv6 zpravy.
							Musime se posunout o IP hlavicku+ICMPv6 hlavicku*/
						icmpRecv6 = (struct icmp6_hdr *) (recvBuffer + sizeof(struct icmp6_hdr) + IPV6_HEADER);

						icmp_idSend46 = &(icmpRecv6->icmp6_dataun.icmp6_un_data16[0]);
						icmp_sequenceSend46 = &(icmpRecv6->icmp6_dataun.icmp6_un_data16[1]);
					}

					/*prekonvertujeme adresu z receiveSock6.sin6_addr do teckove notace IP adresy*/
					inet_ntop(AF_INET6, receiveSock6.sin6_addr.s6_addr, ipAdressString, INET6_ADDRSTRLEN);
#if DEBUG
					printf("ipAdressString(IPv6):%s\n", ipAdressString);
#endif
					/*ziskame jmeno z IP adresy*/
					getHostName(ipAdressString, hostName, true);

					/*Musime zjistit, zda paket, ktery se ztratil, patri nam*/
					lostMyPacketRet = lostMyPacket(recvBuffer, 0, pid, sequence, ttlNull, IPv6);
				}else{
					/*
					ip							icmpRecv4			
					+-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+-+
					/	hlavicka IP protokolu	/	hlavicka ICMP protokolu		/
					+-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+-+-++-+*/

					/*prvni v bufferu bude hlavicka IP protokolu, ip tedy ukazuje na zacatek bufferu:*/
					ip = (struct iphdr *) recvBuffer;

					/*icmpRecv4 ukazuje na ICMP hlavicku. ICMP hlavicka se nachazi za IP hlavickou
					velikost IP hlavicky je v atributu ihl, vydelena ctyrmi*/
					icmpRecv4 = (struct icmphdr *) (recvBuffer + ip->ihl * IHL);

					switch (icmpRecv4->type){
						case ICMP_ECHOREPLY: /*TYPE 0*/
							if (icmpRecv4->code == 0)
								echoReply = true;
							else
								error(ERR_UNKNWN, sockfd, mallocPtr46);				/*otestovano valgrindem*/
							break;
						case ICMP_TIME_EXCEEDED: /*TYPE 11*/
							if (icmpRecv4->code == ICMP_EXC_TTL)	/*TTL = 0*/
								ttlNull = true;
							break;
						case ICMP_DEST_UNREACH:		/*TYPE 3*/
							switch (icmpRecv4->code){
								/*HOST H!*/
								case ICMP_HOST_UNREACH:	/*Host Unreachable*/
								case ICMP_HOST_UNKNOWN: /*Destination Host Unknown*/
								case ICMP_HOST_ISOLATED:/*Source Host Isolated*/
								case ICMP_HOST_ANO:		/*Communication with Destination Host is Administratively Prohibited*/
								case ICMP_HOST_UNR_TOS: /*Destination Host Unreachable for Type of Service*/
								case ICMP_PREC_VIOLATION:/*Host Precedence Violation*/
									/*printf("Prisla zprava HOST.\n");*/
									host = true;
									break;
								/*NETWORK N!*/
								case ICMP_NET_UNKNOWN:	/*Destination Network Unknown*/
								case ICMP_NET_ANO:		/*Communication with Destination Network is Administratively Prohibited*/
								case ICMP_NET_UNR_TOS:	/*Destination Network Unreachable for Type of Service*/
									/*printf("Prisla zprava NETWORK.\n");*/
									network = true;
									break;
								/*Protocol Unreachable P!*/
								case ICMP_PROT_UNREACH: /*Protocol Unreachable*/
									/*printf("Prisla zprava PROTOCOL.\n");*/
									protocol = true;
								break;
								/*communication administratively prohibited X!*/
								case ICMP_PKT_FILTERED:
									/*printf("Prisla zprava COMMUNICATION.\n");*/
									communication = true;
								break;
							}
							break;
					}

					icmp_idSend46 = &(icmpRecv4->un.echo.id);
					icmp_sequenceSend46 = &(icmpRecv4->un.echo.sequence);

					/*prekonvertujeme adresu z receiveSock4.sin_addr do teckove notace IP adresy*/
					inet_ntop(AF_INET, &(receiveSock4.sin_addr), ipAdressString, INET_ADDRSTRLEN );

					getHostName(inet_ntoa(receiveSock4.sin_addr), hostName, false);
					
					/*Musime zjistit, zda paket, ktery se ztratil, patri nam*/
					lostMyPacketRet = lostMyPacket((char *)icmpRecv4, recvLength - ip->ihl * IHL, pid, sequence, ttlNull, IPv6);
				}

				/*spolecne pro IPv4/6*/
				/*VYPIS*/
				if ( lostMyPacketRet ){/*Ztraceny paket patri nam*/
						printf(" %d\t%s (%s) %.3f ms\n", sequence, hostName, ipAdressString, response(timeA, utimeA, timeB, utimeB));
				}
				/*Jedna se o ECHOREPLY, tzn. zajima nas pouze ECHOREPLY typu 0
				naseho pid a nasi sekvence*/
				else if (echoReply &&
					(htons(*icmp_idSend46) == pid) &&
					(htons(*icmp_sequenceSend46) == sequence)){

					printf(" %d\t%s (%s) %.3f ms\n", sequence, hostName,
						 ipAdressString, response(timeA, utimeA, timeB, utimeB));
				}
				else if ( host )
					printf(" %d\t%s (%s) H!\n", sequence, hostName, ipAdressString);
				else if ( network )
					printf(" %d\t%s (%s) N!\n", sequence, hostName, ipAdressString);
				else if ( protocol )
					printf(" %d\t%s (%s) P!\n", sequence, hostName, ipAdressString);
				else if ( communication )
					printf(" %d\t%s (%s) X!\n", sequence, hostName, ipAdressString);

			}
			else{/*TIMEOUT, pokud neprisel pozadovany paket do 3sekund, vypiseme * dle zadani*/
				printf(" %d\t* * *\n", sequence);
				break;/*prerusime cyklus*/
			}

		/*aby podminka while pod timto ifem nebyla slozita a neprehledna, nektere podminky specifikuji uz zde:*/
		if(host || network || protocol || communication)
			break; /*koncime*/

		/*Vyskoc pokud:
		dorazi ztraceny paket ||
		dorazi ECHOREPLY && ID==pid && sequence==sequence
		*/
		}while(! (lostMyPacketRet || ( echoReply && (ntohs(*icmp_idSend46) == pid) && (ntohs(*icmp_sequenceSend46) == sequence) ) ));

		++sequence;
		++param.firstTTL;

	/*aby podminka while pod timto ifem nebyla slozita a neprehledna, nektere podminky specifikuji uz zde:*/
	if(host || network || protocol || communication)
		break; /*koncime*/

	/*Podminky pro odeslani paketu:
	Odesilame ICMP pakety dokud TTL nenabude max. hodnoty nebo dokud neprijde
	ECHOREPLY pro NAS pozadavek(tzn ID, SEQUENCE)*/

	} while ( (param.firstTTL != param.maxTTL+1) && /*+1 je aby se provedl i ten posledni*/
			(!(echoReply &&
			((ntohs(*icmp_idSend46)) == pid) &&
			((ntohs(*icmp_sequenceSend46)) == (sequence-1)))));


	close(sockfd);
	sockfd = -1;
	free(mallocPtr46);
	mallocPtr46 = NULL;
}

bool setTTL(int sockfd, int ttl, bool IPv6){
/*Nastavuje TTL na soketu, osetri chybu.*/

	int setsockoptReturnVal = 0;/*navratova hodnota funkce setsockopt*/
	
	/*nastaveni TTL pomoci funkce setsockopt-nastavit moznost na socketu
		SOL_IP		vlasnost IP protokolu
		IP_TTL/IPV6_UNICAST_HOPS	pocet smerovacu, kterym muze paket projit
		ttl			hodnota
		sizeof(ttl)	delka
	*/
	if(IPv6)
		setsockoptReturnVal = setsockopt(sockfd, SOL_IPV6, IPV6_UNICAST_HOPS,(const char *)&ttl, sizeof(ttl));
	else
		setsockoptReturnVal = setsockopt(sockfd, SOL_IP, IP_TTL,(const char *)&ttl, sizeof(ttl));

	if (setsockoptReturnVal == -1){/*chyba setsockopt*/
		perror("Chyba funkce setsockopt():");
		return false;
	}

	return true;
}

void getParams(int argc, char* argv[], Tparameter *param){
/*Zpracuje parametry, osetri chybove stavy a ulozi parametry do struktury.*/

	int c; 			/*navrat. hodnota getopt*/
	int sscanfB; 	/*navratova hodnota sscanf*/

	if(argc < MIN_PARAM || argc > MAX_PARAM)/*spatne zadane parametry*/
		error(ERR_PARAM, false, NULL);
	else if(argc == 2 && ((strcmp("-h", argv[1])) == 0)){/*napoveda*/
		printf("%s", help);
		exit(0);
	}

	while ( (c = getopt(argc, argv, "f:m:" )) != -1 ){

		sscanfB = 0;

		switch(c){
			case 'f':
				/*printf("Prepinac:\t-f\ts parametrem:\t%s\n", optarg);*/
				/*Konverze retezce na cislo*/
				sscanfB = sscanf(optarg, "%d", &(param->firstTTL));
				/*Nepovedla se konverze retezce na cislo*/
				if ( sscanfB == 0 || sscanfB == EOF )
					error(ERR_SSCANF, false, NULL);

				break;
			case 'm':
				/*printf("Prepinac:\t-m\ts parametrem:\t%s\n", optarg);*/
				/*Konverze retezce na cislo*/
				sscanfB = sscanf(optarg, "%d", &(param->maxTTL));

				/*Nepovedla se konverze retezce na cislo*/
				if ( sscanfB == 0 || sscanfB == EOF )
					error(ERR_SSCANF, false, NULL);

				break;
			case '?':
				error(ERR_PARAM, false, NULL);
		}

	}

	/*kontrola host parametru*/
	if(argv[optind] == NULL)/*ocekava se jeste jeden parametr ktery nebyl zadan*/
		error(ERR_PARAM, false, NULL);

	/*kontrola firstTTL a maxTTL*/
	if(param->firstTTL < 1 || param->maxTTL < 1)
		error(ERR_TTL, false, NULL);
	else if( param->maxTTL < param->firstTTL ){
		printf("Nelze dosahnout cile.\n");
		exit(0);
	}

	param->host = argv[optind];
}

void error(int number, int sockfd, void* malloc){
/*Muze/nemusi vytisknout chybu
Muze/nemusi uvolnit pamet
a ukonci program.

number 		je cislo chyby, v pripade -1 se nic nevypisuje, (perror)
sockfd 		je popisovac souboru, pokud tuhle funkci zavola prikaz, ktery sockfd 
nevyuzije, za sockfd dosadi 0(STDIN) cemuz nevyhovi podminka IF.
*/
	if(number >= 0)
		fprintf(stderr, "%s", errorMessages[number]);

	if(sockfd != -1)/*chceme zavrit socket*/
		if(close(sockfd) == -1)	/*zavrit socket se nepovede*/
			perror("Chyba funkce close():");

	if(malloc != NULL)
		free(malloc);

	exit(-1);
}
																	
/*Funkce Checksum prevzata z RFC 1071 - Computing the Internet
 checksum bod 4.1 "C"
URL: www.tools.ietf.org/rfc/rfc1071.txt
*/
unsigned short checksum(unsigned short *ip, int len){
/*Spocita Checksum, ktery se doplni do ICMPv4/v6 hlavicky.*/

		register long sum = 0;

		while( len > 1 )  {
		/*  This is the inner loop */
			sum += *ip++;
			len -= 2;
		}

		 /*  Add left-over byte, if any */
		if( len > 0 )
			sum += * (unsigned char *) ip;

		/*  Fold 32-bit sum to 16 bits */
		while (sum>>16)
			sum = (sum & 0xffff) + (sum >> 16);

	return ~sum;
}

bool lostMyPacket(char* recvBuffer, unsigned short ipHeaderLength, unsigned short id, unsigned short sequence, bool ttlNull, bool IPv6){
/*Funkce zjisti, zda se jedna o nas ztraceny paket IPv4/IPv6.
Pokud ztraceny paket je nas vraci TRUE, jinak FALSE*/

	if (IPv6){
		struct icmp6_hdr *icmpSend6;

		icmpSend6 = (struct icmp6_hdr *) (recvBuffer + sizeof(struct icmp6_hdr) + IPV6_HEADER);
#if DEBUG
		printf("TYPE ve ztracenem paketu ICMPv6:%d\n", icmpSend6->icmp6_type);
		printf("CODE ve ztracenem paketu ICMPv6:%d\n", icmpSend6->icmp6_code);
		printf("PID ve ztracenem paketu ICMPv6:%d == %d\n", ntohs(icmpSend6->icmp6_dataun.icmp6_un_data16[0]), id);
		printf("SEQ ve ztracenem paketu ICMPv6:%d == %d\n", ntohs(icmpSend6->icmp6_dataun.icmp6_un_data16[1]), sequence);
#endif

		if ( ttlNull &&
			 icmpSend6->icmp6_type == ICMP6_ECHO_REQUEST &&
			 icmpSend6->icmp6_code == 0 &&
			 ntohs(icmpSend6->icmp6_dataun.icmp6_un_data16[0]) == id &&
			 ntohs(icmpSend6->icmp6_dataun.icmp6_un_data16[1]) == sequence)
			return true;
		else
			return false;
		
	}else{
		struct iphdr *ipSend;
		struct icmphdr *icmpSend4;

		/*Alokovany buffer je prilis maly. POZOR: hlavicka muze byt vetsi a tato
		podminka muze projit. Proto konrolujeme velikost jeste jednou.*/
		if ( ipHeaderLength < ( sizeof(struct iphdr) + (2 * sizeof(struct icmphdr)) ) )
			return false;

		/*ukazatel na IP hlavicku*/
		ipSend = (struct iphdr *)(recvBuffer + sizeof(struct icmphdr));

		/*Velikost buffer je stale mala*/
		if (ipHeaderLength < (2 * sizeof(struct icmphdr)) + ipSend->ihl * IHL)
			return false;

		/*ihl-obsahuje delku hlavičky IP paketu vydelenou ctyrmi.*/
		icmpSend4 = (struct icmphdr *) ((char *)ipSend + ipSend->ihl * IHL);
#if DEBUG
		printf("ICMPv4:icmpSend->type:%d\n", ntohs(icmpSend4->type));
		printf("ICMPv4:icmpSend->code:%d\n", ntohs(icmpSend4->code));
		printf("ICMPv4:icmpSend->un.echo.id:%d\n", ntohs(icmpSend4->un.echo.id));
		printf("ICMPv4:icmpSend->un.echo.sequence:%d\n", ntohs(icmpSend4->un.echo.sequence));
#endif

		/*Pokud to ma byt nas ztraceny paket musi:
		vyprset TTL, promenna ttlNull
		muselo byt poslane ECHO CODE 0, musi souhlasit ID a sekvence*/
		if ( ttlNull &&
			(icmpSend4->type == ICMP_ECHO) && (icmpSend4->code == 0) &&
			(ntohs(icmpSend4->un.echo.id) == id) && (ntohs(icmpSend4->un.echo.sequence) == sequence))
			return true;
		else
			return false;
	}

	return false;
}

